/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file daisylibao.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 15.02.2006
 *
 * An interface against libao. Audio decoders should use this instead of using libao directly.
 */

#include <ao/ao.h>
#include <stdio.h>
#include <stdbool.h>
#include <errno.h>
#include <assert.h>

#include "daisylibao.h"
#include "common.h"
#include "audio.h"
#include "report.h"
#include "snprintf/snprintf.h"



/* ************************************************** *
 * 			Static function declarations 			  *
 * ************************************************** */

static
void ao_open_live_error (void);



/* ************************************************** *
 * 			Static function definitions 			  *
 * ************************************************** */

/**
 * This function is called if ao_open_live returns NULL. This means we have
 * an error, and this method reports it based on errno
 */
static
void ao_open_live_error (void)
{
	switch (errno)
	{
		case AO_ENODRIVER:
			report ("No driver corresponts to driver ID.", REP_ERROR);
			break;
		case AO_ENOTLIVE:
			report ("This driver is not a live output device.", REP_ERROR);
			break;
		case AO_EBADOPTION:
			report ("A valid option key has an invalid value.", REP_ERROR);
			break;
		case AO_EOPENDEVICE:
			report ("Cannot open the device.", REP_ERROR);
			break;
		case AO_EFAIL:
			report ("Unknown reason.", REP_ERROR);
			break;
		default:
			report ("ao_open_live_error called with unknown error", REP_ERROR);
			break;
	}
}



/* ************************************************** *
 * 			Global function definitions 			  *
 * ************************************************** */

/**
 * Initialises the libao-part of the audio engine.
 * Must be called before the libao part of the audio engine is to be used.
 * @param int byte_format - the byte format of the audio samples to play.
 * @param int channels - channels of the audio to be played.
 * @param int samplerate - samplerate of the audio to be played.
 * @return ao_device * - returns a pointer to the aodevice that should be used in playback.
 */
ao_device * libao_initiate (int byte_format, int channels, int samplerate)
{
	ao_device *device = NULL;
	
	ao_sample_format format;
	int libao_driver_id = -1;

	ao_initialize ();
	libao_driver_id =  ao_default_driver_id ();
	if (libao_driver_id == -1) {
		report("Could not get a valid driver ID", REP_ERROR);
		return false;
	}
	format.bits = 16; /* Do decoder always give us 16bits? */
/*	format.byte_format = AO_FMT_LITTLE; */
	format.byte_format = byte_format;
	format.channels = channels;
	format.rate = samplerate;

	device = ao_open_live (libao_driver_id, &format, NULL);
	if (device == NULL)
	{
		ao_open_live_error();
		return NULL;
	}
	else if (DEBUG_LIBAO)
	{
		char msg[STRLEN];
		report ("libao initialized", REP_DEBUG);
		snprintf (msg, STRLEN, "libao info:\n\tbyte_format: %d\n\tchannels: %d\n\tsamplerate: %d", byte_format, channels, samplerate);
		report (msg, REP_DEBUG);
	}
	return device;
}

/**
 * Outputs samples from a buffer. (basically wraps the ao_play() function).
 * @param ao_device *device - output device to play to.
 * @param void *output_samples - number of samples to play.
 * @param int num_bytes - number of bytes to play.
 * @return int - returns 0 on success, anything else on fail.
 */
int libao_play (ao_device *device, void *output_samples, int num_bytes)
{
	return ao_play(device, output_samples, num_bytes);
}

/**
 * Terminates libao part of audio engine the Proper Way [tm]
 * @param ao_device *device - the aodevice to terminate.
 */
void libao_terminate (ao_device *device)
{
	ao_close (device);
	ao_shutdown ();
	if (DEBUG_LIBAO) report ("libao closed", REP_DEBUG);
}
